from __future__ import annotations

import logging
from dataclasses import asdict, dataclass
from functools import partial
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    List,
    Literal,
    Optional,
    Set,
    Union,
    get_args,
)

from polars.plugins import register_plugin_function

from polars_ols.utils import build_expressions_from_patsy_formula, parse_into_expr

if TYPE_CHECKING:
    import polars as pl
    from polars.type_aliases import IntoExpr

    ExprOrStr = Union[pl.Expr, str]

logger = logging.getLogger(__name__)

__all__ = [
    # fitting and prediction functions
    "compute_least_squares",
    "compute_recursive_least_squares",
    "compute_rolling_least_squares",
    "compute_least_squares_from_formula",
    "compute_multi_target_least_squares",
    "predict",
    # model specific parameters
    "OLSKwargs",
    "RLSKwargs",
    "RollingKwargs",
    # types controlling general modelling behaviour
    "NullPolicy",
    "OutputMode",
    "SolveMethod",
]

NullPolicy = Literal[
    "zero",  # simply zero fills nulls in both targets & features
    "drop",  # drops any rows with nulls in fitting and masks associated predictions with nulls
    "ignore",  # use this option if nulls are already handled upstream
    "drop_zero",  # drops any rows with nulls in fitting, but then computes predictions
    # with zero filled features. Use this to allow for extrapolation.
    "drop_y_zero_x",  # only drops rows with null targets and fill any null features with zero
    "drop_window",  # only relevant to rolling window regression: this causes any observations
    # with nulls to be omitted and only valid observations within the fixed window are used.
]
OutputMode = Literal["predictions", "residuals", "coefficients"]
SolveMethod = Literal["qr", "svd", "chol", "lu", "cd", "cd_active_set"]

_VALID_NULL_POLICIES: Set[NullPolicy] = set(get_args(NullPolicy))
_VALID_OUTPUT_MODES: Set[OutputMode] = set(get_args(OutputMode))
_VALID_SOLVE_METHODS: Set[SolveMethod] = set(get_args(SolveMethod)).union({None})


@dataclass
class Kwargs:
    null_policy: NullPolicy = "ignore"

    def to_dict(self) -> Dict[str, Any]:
        return asdict(self)

    def __post_init__(self):
        # rust code does validate all options, but prefer to fail, on some, early
        assert (
            self.null_policy in _VALID_NULL_POLICIES
        ), f"'null_policy' must be one of {_VALID_NULL_POLICIES}. You passed: {self.null_policy}"


@dataclass
class OLSKwargs(Kwargs):
    """Specifies parameters relevant for regularized linear models: LASSO / Ridge / ElasticNet.

    Attributes:
        alpha: Regularization strength. Defaults to 0.0.
        l1_ratio: Mixing parameter for ElasticNet regularization (0 for Ridge, 1 for LASSO).
            Defaults to None (equivalent to Ridge regression).
        max_iter: Maximum number of iterations. Defaults to 1000 iterations.
        tol: Tolerance for convergence criterion. Defaults to 1.e-5.
        positive: Whether to enforce non-negativity constraints on coefficients.
            Defaults to False (no constraint on coefficients).
        null_policy: Strategy for handling missing data. Defaults to "ignore".
        solve_method: Algorithm used for computing least squares solution.
            Defaults to None, where a recommended default method is chosen based on problem
            specifics.
        rcond: Optional float specifying cut-off ratio for small singular values. Only relevant for
               "SVD" solve methods. Defaults to None, where it is chosen as per
                numpy lstsq convention.
    """

    alpha: Optional[float] = 0.0
    l1_ratio: Optional[float] = None
    max_iter: Optional[int] = 1_000
    tol: Optional[float] = 1.0e-5
    positive: Optional[bool] = False  # if True, imposes non-negativity constraint on coefficients
    solve_method: Optional[SolveMethod] = None
    rcond: Optional[float] = None

    def __post_init__(self):
        # rust code does validate all options, but prefer to fail, on some, early
        valid_ols_policies = _VALID_NULL_POLICIES - {"drop_window"}
        # 'drop_window' is specific to rolling window models.
        assert (
            self.null_policy in valid_ols_policies
        ), f"'null_policy' must be one of {valid_ols_policies}. You passed: {self.null_policy}"
        assert (
            self.solve_method in _VALID_SOLVE_METHODS
        ), f"'solve_method' must be one of {_VALID_SOLVE_METHODS}. You passed: {self.solve_method}"


@dataclass
class RLSKwargs(Kwargs):
    """Specifies parameters of Recursive Least Squares models.

    Attributes:
        half_life: Half-life parameter for exponential forgetting. Defaults to None, which is
                          equivalent to expanding window least-squares (no forgetting).
        initial_state_mean: Denotes a prior for the mean value of the coefficients.
                          It can be either a scalar or a list. Defaults to None (no prior).
        initial_state_covariance: Scalar which denotes a prior of the uncertainty (variance)
                          around `initial_state_mean`. It can be thought of as inversely
                          proportional to the strength of an L2 penalty in a ridge regression.
                          Defaults to 10, intended to be a weak 'diffuse' prior.
        null_policy: Strategy for handling missing data. Defaults to "drop".
    """

    half_life: Optional[float] = None
    initial_state_covariance: Optional[float] = 10.0
    initial_state_mean: Union[Optional[List[float], float]] = None
    null_policy: NullPolicy = "drop"


@dataclass
class RollingKwargs(Kwargs):
    """Specifies parameters of Rolling OLS model.

    Attributes:
        window_size: The size of the rolling window.
        min_periods: The minimum number of observations required to produce estimates.
        use_woodbury: Whether to use Woodbury matrix identity for faster computation.
                      Defaults to True if num_features > 10.
        alpha: L2 Regularization strength. Default is 0.0.
        null_policy: Strategy for handling missing data. Defaults to "drop_window".
    """

    window_size: int = 1_000_000  # defaults to expanding OLS
    min_periods: Optional[int] = None
    use_woodbury: Optional[bool] = None
    alpha: Optional[float] = None  # optional ridge alpha
    null_policy: NullPolicy = "drop_window"


def _pre_process_data(
    target: ExprOrStr,
    *features: ExprOrStr,
    sample_weights: Optional[ExprOrStr],
    add_intercept: bool,
):
    """Pre-processes the input data by adding an intercept to features and applying sample weights,
    if specified.

    Args:
        target: The target expression.
        *features: Variable number of feature expressions.
        sample_weights: Optional expression representing sample weights.
        add_intercept: Whether to add an intercept column.

    Returns:
        Tuple containing the pre-processed target, features, and sample weights.
    """
    target = parse_into_expr(target)
    features = [parse_into_expr(f) for f in features]
    # handle intercept
    if add_intercept:
        if any(f.meta.output_name == "const" for f in features):
            logger.info("feature named 'const' already detected, assuming it is an intercept")
        else:
            features.append(target.fill_null(0.0).mul(0.0).add(1.0).alias("const"))
    # handle sample weights
    sqrt_w: Optional[pl.Expr] = None
    if sample_weights is not None:
        sqrt_w = parse_into_expr(sample_weights).sqrt()
        target *= sqrt_w
        features = [(expr * sqrt_w) for expr in features]
    return target, features, sqrt_w


def _register_least_squares_plugin(
    target: ExprOrStr,
    *features: ExprOrStr,
    mode: OutputMode,
    function_name: str,
    ols_kwargs: Kwargs,
    returns_scalar_coefficients: bool = False,
    **kwargs,
):
    # pre-process features and targets prior to fitting
    target = parse_into_expr(target)
    target_fit, features_fit, sqrt_w = _pre_process_data(target, *features, **kwargs)

    # register either coefficient or prediction plugin functions
    if mode == "coefficients":
        # TODO: remove 'rename_fields' after https://github.com/pola-rs/pyo3-polars/issues/79
        #  it currently breaks input_wildcard_expansion=True correctly returning a struct.
        return (
            register_plugin_function(
                plugin_path=Path(__file__).parent,
                function_name=f"{function_name}_coefficients",
                args=[target_fit, *features_fit],
                kwargs=ols_kwargs.to_dict(),
                is_elementwise=False,
                changes_length=returns_scalar_coefficients,
                returns_scalar=returns_scalar_coefficients,
                input_wildcard_expansion=True,
            )
            .alias("coefficients")
            .struct.rename_fields([f.meta.output_name() for f in features_fit])
        )
    else:
        predictions = register_plugin_function(
            plugin_path=Path(__file__).parent,
            function_name=function_name,
            args=[target_fit, *features_fit],
            kwargs=ols_kwargs.to_dict(),
            is_elementwise=False,
            input_wildcard_expansion=True,
        )
        if sqrt_w is not None:
            predictions /= sqrt_w  # undo the scaling implicit in WLS weighting
        if mode == "predictions":
            return predictions
        else:
            return target - predictions


def compute_least_squares(
    target: ExprOrStr,
    *features: ExprOrStr,
    sample_weights: Optional[ExprOrStr] = None,
    add_intercept: bool = False,
    mode: OutputMode = "predictions",
    ols_kwargs: Optional[OLSKwargs] = None,
) -> pl.Expr:
    """Performs least squares regression.

    Depending on parameters provided this method supports a combination of sample weighting (WLS),
     L1/L2 regularization,
     and/or non-negativity constraint on coefficients.

    Args:
        target: The target expression.
        *features: Variable number of feature expressions.
        sample_weights: Optional expression representing sample weights.
        add_intercept: Whether to add an intercept column.
        mode: Mode of operation ("predictions", "residuals", "coefficients").
        ols_kwargs: Additional keyword arguments specific for regularized OLS models. See OLSKwargs.
    Returns:
        Resulting expression based on the chosen mode.
    """
    assert mode in _VALID_OUTPUT_MODES, f"'mode' must be one of {_VALID_OUTPUT_MODES}"
    ols_kwargs: OLSKwargs = ols_kwargs or OLSKwargs()

    # register either coefficient or prediction plugin functions
    return _register_least_squares_plugin(
        target,
        *features,
        mode=mode,
        function_name="least_squares",
        ols_kwargs=ols_kwargs,
        sample_weights=sample_weights,
        add_intercept=add_intercept,
        returns_scalar_coefficients=True,
    )


def compute_multi_target_least_squares(
    targets: ExprOrStr,
    *features: ExprOrStr,
    sample_weights: Optional[ExprOrStr] = None,
    add_intercept: bool = False,
    mode: OutputMode = "predictions",
    ols_kwargs: Optional[OLSKwargs] = None,
) -> pl.Expr:
    """Performs multi-target least squares regression.

    This function expects 'targets' to be of type struct with field names corresponding
    to each target series name. It returns an expression yielding a struct with similar field names
    and either predictions or residuals as values depending on the chosen mode.

    This is more efficient, but equivalent to, running 'N' multiple linear regressions of each
    underlying target onto the same set of features.

    OLS, WLS, & Ridge for both "prediction" and "residuals" modes are currently supported.
    For other configurations (e.g. LASSO): consider running multiple independent regressions on
    a multi-expression "target" using 'compute_least_squares'.
    """
    ols_kwargs = ols_kwargs or OLSKwargs()
    multi_target_conditions = not ols_kwargs.positive and (
        ols_kwargs.l1_ratio is None or ols_kwargs.l1_ratio == 0.0
    )
    msg = "Consider running multiple independent regressions on a multi-expression target!"
    assert multi_target_conditions, (
        "Multi-target regression is only supported " "for unconstrained OLS & Ridge problems." + msg
    )
    assert ols_kwargs.solve_method in {"svd", None}, (
        "only solve_method='svd' is supported for " "multi-target regressions"
    )
    if mode == "coefficients":
        raise NotImplementedError(
            "Only mode={'predictions', 'residuals'} " "is currently supported. " + msg
        )

    return _register_least_squares_plugin(
        targets,
        *features,
        function_name="multi_target_least_squares",
        ols_kwargs=ols_kwargs,
        sample_weights=sample_weights,
        add_intercept=add_intercept,
        returns_scalar_coefficients=True,
        mode=mode,
    )


def compute_recursive_least_squares(
    target: ExprOrStr,
    *features: ExprOrStr,
    sample_weights: Optional[ExprOrStr] = None,
    add_intercept: bool = False,
    mode: OutputMode = "predictions",
    rls_kwargs: Optional[RLSKwargs] = None,
) -> pl.Expr:
    """Performs an efficient recursive least squares regression (RLS).

    Defaults to RLS with forgetting factor of 1.0 and a high initial state variance: equivalent to
     efficient 'streaming' expanding window OLS.

    Args:
        target: The target expression.
        *features: Variable number of feature expressions.
        sample_weights: Optional expression representing sample weights.
        add_intercept: Whether to add an intercept column.
        mode: Mode of operation ("predictions", "residuals", "coefficients").
        rls_kwargs: Additional keyword arguments for the recursive least squares model.
                    See RLSKwargs.

    Returns:
        Resulting expression based on the chosen mode.
    """
    assert mode in _VALID_OUTPUT_MODES, f"'mode' must be one of {_VALID_OUTPUT_MODES}"
    rls_kwargs: RLSKwargs = rls_kwargs or RLSKwargs()
    # register either coefficient or prediction plugin functions
    return _register_least_squares_plugin(
        target,
        *features,
        mode=mode,
        function_name="recursive_least_squares",
        ols_kwargs=rls_kwargs,
        sample_weights=sample_weights,
        add_intercept=add_intercept,
    )


def compute_rolling_least_squares(
    target: ExprOrStr,
    *features: ExprOrStr,
    sample_weights: Optional[ExprOrStr] = None,
    add_intercept: bool = False,
    mode: OutputMode = "predictions",
    rolling_kwargs: Optional[RollingKwargs] = None,
) -> pl.Expr:
    """Performs least squares regression in a rolling window fashion.

    Args:
        target: The target expression.
        *features: Variable number of feature expressions.
        sample_weights: Optional expression representing sample weights.
        add_intercept: Whether to add an intercept column.
        mode: Mode of operation ("predictions", "residuals", "coefficients").
        rolling_kwargs: Additional keyword arguments for the rolling least squares model.
                        See RollingKwargs.

    Returns:
        Resulting expression based on the chosen mode.
    """
    assert mode in _VALID_OUTPUT_MODES, f"'mode' must be one of {_VALID_OUTPUT_MODES}"
    rolling_kwargs: RollingKwargs = rolling_kwargs or RollingKwargs()
    # register either coefficient or prediction plugin functions
    expr = _register_least_squares_plugin(
        target,
        *features,
        mode=mode,
        function_name="rolling_least_squares",
        ols_kwargs=rolling_kwargs,
        sample_weights=sample_weights,
        add_intercept=add_intercept,
    )
    if mode in {"predictions", "residuals"}:
        expr = expr.fill_nan(None)
    return expr


def compute_least_squares_from_formula(
    formula: str,
    sample_weights: Optional[ExprOrStr] = None,
    mode: OutputMode = "predictions",
    **kwargs,
) -> pl.Expr:
    """Performs least squares regression using a formula.

    Depending on choice of additional kwargs dispatches either rolling, recursive,
    on static least squares compute functions.

    Args:
        formula: Patsy-style formula string.
        sample_weights: Optional expression representing sample weights.
        mode: Mode of operation ("predictions", "residuals", "coefficients").
        **kwargs: Additional keyword arguments for the least squares function.

    Returns:
        Resulting expression based on the formula.
    """
    expressions, add_intercept = build_expressions_from_patsy_formula(
        formula, include_dependent_variable=True
    )
    # resolve additional kwargs and relevant ols compute function
    if kwargs.get("half_life"):
        rls_kwargs: RLSKwargs = RLSKwargs(**kwargs)
        func = partial(compute_recursive_least_squares, rls_kwargs=rls_kwargs)
    elif kwargs.get("window_size"):
        rolling_kwargs: RollingKwargs = RollingKwargs(**kwargs)
        func = partial(compute_rolling_least_squares, rolling_kwargs=rolling_kwargs)
    else:
        ols_kwargs: OLSKwargs = OLSKwargs(**kwargs)
        func = partial(compute_least_squares, ols_kwargs=ols_kwargs)

    return func(
        expressions[0],
        *expressions[1:],
        add_intercept=add_intercept,
        sample_weights=sample_weights,
        mode=mode,
    )


def predict(
    coefficients: ExprOrStr,
    *features: ExprOrStr,
    null_policy: NullPolicy = "zero",
    add_intercept: bool = False,
    name: Optional[str] = None,
) -> pl.Expr:
    """Helper which computes predictions as a product of (aligned) coefficients with features.

    Args:
        coefficients: Polars expression returning a coefficients struct.
        *features: variable number of feature expressions.
        null_policy: specifies how nulls in features are handled. Defaults to zero filling.
        add_intercept: boolean indicating if a constant should be added to features.
        name: optional str defining an alias for computed predictions expression.

    Returns:
        polars expression denoting computed predictions.
    """
    assert null_policy in _VALID_NULL_POLICIES, "'null_policy' must be one of {drop, ignore, zero}"

    coefficients: pl.Expr = parse_into_expr(coefficients)
    features: List[pl.Expr] = [parse_into_expr(f) for f in features]

    if add_intercept:
        if any(f.meta.output_name == "const" for f in features):
            logger.warning("feature named 'const' already detected, assuming it is the intercept")
        else:
            features += (features[-1].fill_null(0.0).mul(0.0).add(1.0).alias("const"),)
    return register_plugin_function(
        plugin_path=Path(__file__).parent,
        function_name="predict",
        args=[coefficients, *features],
        kwargs={"null_policy": null_policy},
        is_elementwise=False,
        input_wildcard_expansion=True,
    ).alias(name or "predictions")



from typing import Optional, Tuple

import numpy as np
import polars as pl
import pytest
import statsmodels.formula.api as smf
from sklearn.linear_model import ElasticNet, Ridge
from statsmodels.regression.rolling import RollingOLS

from polars_ols import (
    NullPolicy,
    OLSKwargs,
    SolveMethod,
    compute_least_squares,
    compute_least_squares_from_formula,
    compute_multi_target_least_squares,
)
from polars_ols.utils import timer


def _make_data(
    n_samples: int = 5_000,
    n_features: int = 2,
    n_groups: Optional[int] = None,
    scale: float = 0.1,
    sparsity: float = 0.0,
    add_missing: bool = False,
    missing_columns: Optional[Tuple[str, ...]] = None,
) -> pl.DataFrame:
    rng = np.random.default_rng(0)
    x = rng.normal(size=(n_samples, n_features))
    eps = rng.normal(size=n_samples, scale=scale)

    df = pl.DataFrame(data=x, schema=[f"x{i + 1}" for i in range(n_features)]).with_columns(
        y=pl.lit(x[:, : int(n_features * (1.0 - sparsity))].sum(1) + eps)
    )

    if n_groups is not None:
        df = df.with_columns(group=pl.lit(rng.integers(n_groups, size=n_samples)))

    if add_missing:

        def insert_nulls(val):
            return None if rng.random() < 0.1 else val

        columns = missing_columns or df.columns
        df = df.with_columns(
            *(pl.col(c).map_elements(insert_nulls, return_dtype=pl.Float64) for c in columns)
        )
    return df


@pytest.mark.parametrize("solve_method", ("qr", "svd", "chol", "lu", None))
def test_ols(solve_method: SolveMethod):
    import os

    os.environ["POLARS_VERBOSE"] = "1"
    df = _make_data(n_samples=1_000, n_features=2)
    # compute OLS w/ polars-ols
    with timer(f"\nOLS {solve_method}", precision=5):
        for _ in range(1_000):
            expr = compute_least_squares(
                pl.col("y"),
                pl.col("x1"),
                pl.col("x2"),
                ols_kwargs=OLSKwargs(solve_method=solve_method),
            ).alias("predictions")
            df = df.lazy().with_columns(expr).collect()
    # compute OLS w/ lstsq numpy
    with timer("OLS numpy", precision=5):
        for _ in range(1_000):
            x, y = df.select("x1", "x2").to_numpy(), df.select("y").to_numpy().flatten()
            coef = np.linalg.lstsq(x, y, rcond=None)[0]
            df = df.with_columns(predictions2=pl.lit(x @ coef).flatten())
    assert np.allclose(df["predictions"], df["predictions2"], atol=1.0e-4, rtol=1.0e-4)


@pytest.mark.parametrize(
    "alpha,mode,null_policy",
    [
        (0.0, "residuals", "ignore"),
        (0.0, "residuals", "drop"),
        (0.0001, "residuals", "drop_y_zero_x"),
        (0.01, "residuals", "drop_zero"),
    ],
)
def test_multi_target_regression(alpha, mode, null_policy):
    df = _make_data(
        n_samples=10_000,
        n_features=3,
        add_missing=null_policy not in {"zero", "ignore"},
        missing_columns=("x1",),
    )
    df = df.with_columns(
        pl.struct(
            y1=pl.col("x1") + pl.col("x2") + pl.col("x3"),
            y2=pl.col("x1") - pl.col("x2") + pl.col("x3"),
            y3=-pl.col("x1") + pl.col("x2") - pl.col("x3"),
        ).alias("y")
    )
    ols_kwargs = OLSKwargs(null_policy=null_policy, solve_method="svd", alpha=alpha)

    with timer(f"compute multi-target (alpha={alpha}, null_policy={null_policy}, mode={mode})"):
        multi_target = df.select(
            compute_multi_target_least_squares(
                "y",
                "x1",
                "x2",
                "x3",
                mode=mode,
                ols_kwargs=ols_kwargs,
            ).alias(mode)
        )

    with timer("compute multiple linear regressions"):
        expected = df.unnest("y").select(
            compute_least_squares(
                target,
                "x1",
                "x2",
                "x3",
                mode=mode,
                ols_kwargs=ols_kwargs,
            ).alias(target)
            for target in ("y1", "y2", "y3")
        )

    assert np.allclose(multi_target.unnest(mode), expected, equal_nan=True)


def test_fit_missing_data_coefficients():
    df = _make_data(add_missing=True)

    # # in presence of unhandled nulls assert the rust library raises ComputeError
    # with pytest.raises(pl.exceptions.ComputeError):
    #     df.select(
    #         pl.col("y").least_squares.ols(
    #             pl.col("x1"), pl.col("x2"), null_policy="ignore", mode="coefficients"
    #         )
    #     )

    # test rust zero policy is sane
    assert np.allclose(
        df.select(
            pl.col("y").least_squares.ols(
                pl.col("x1"), pl.col("x2"), null_policy="zero", mode="coefficients"
            )
        ).unnest("coefficients"),
        df.fill_null(0.0)
        .select(
            pl.col("y").least_squares.ols(
                pl.col("x1"), pl.col("x2"), null_policy="ignore", mode="coefficients"
            )
        )
        .unnest("coefficients"),
    )

    # test rust drop (any) policy is sane
    assert np.allclose(
        df.select(
            pl.col("y").least_squares.ols(
                pl.col("x1"), pl.col("x2"), null_policy="drop", mode="coefficients"
            )
        ).unnest("coefficients"),
        df.drop_nulls()
        .select(
            pl.col("y").least_squares.ols(
                pl.col("x1"), pl.col("x2"), null_policy="ignore", mode="coefficients"
            )
        )
        .unnest("coefficients"),
    )

    # test rust drop_y_zero_x policy is sane
    assert np.allclose(
        df.select(
            pl.col("y").least_squares.ols(
                pl.col("x1"), pl.col("x2"), null_policy="drop_y_zero_x", mode="coefficients"
            )
        ).unnest("coefficients"),
        df.drop_nulls(subset=["y"])
        .fill_null(0.0)
        .select(
            pl.col("y").least_squares.ols(
                pl.col("x1"), pl.col("x2"), null_policy="ignore", mode="coefficients"
            )
        )
        .unnest("coefficients"),
    )


@pytest.mark.parametrize("null_policy", ["drop", "drop_zero", "drop_y_zero_x"])
def test_fit_missing_data_predictions_and_residuals(null_policy: NullPolicy):
    df = _make_data(add_missing=True)

    # check predictions logic
    with timer("numpy lstsq w/ manual nan policy", precision=5):
        # compute desired behaviour for "drop" in numpy:
        x, y = (
            df.select("x1", "x2").to_numpy(),
            df.select("y").to_numpy().flatten(),
        )  # notice implicit null -> nan converting to numpy

        # 1) compute mask, drop invalid rows, compute coefficients
        if null_policy == "drop_y_zero_x":
            # drop_y will only drop rows based on the target
            is_valid = ~np.isnan(y)
            x_fit, y_fit = np.nan_to_num(x[is_valid, :]), y[is_valid]
        else:
            # drop or drop_zero will drop any invalid row in fitting
            is_valid = ~np.isnan(x).any(axis=1) & ~np.isnan(y)
            x_fit, y_fit = x[is_valid, :], y[is_valid]

        coef = np.linalg.lstsq(x_fit, y_fit, rcond=None)[0]

        # in order to broadcast (valid) predictions/residuals to the dimensions of original data; we
        # use the original (un-dropped) x for computing predictions.
        # The behaviours which are provided for handling nulls in prediction context:
        # a) null_policy="drop": ensure predictions are always masked to null
        #                        if any feature or target is null.
        # b) null_policy="drop_zero": using coefficients estimated from dropped data, always produce
        #                             valid predictions by zero filling features
        #                             (this allows extrapolation).
        # c) null_policy="drop_y_zero_x": similar to "drop_zero" but where coefficients have been
        #                                 estimated based on the validity of y alone.
        is_nan_x = np.isnan(x)
        x_predict = x.copy()
        x_predict[is_nan_x] = 0.0  # fill x nans with zero
        predictions_expected = x_predict @ coef

        # if user selected "drop" ensure that predictions are masked based on
        # validity of both features & targets.
        if null_policy == "drop":
            predictions_expected[~is_valid] = np.nan
        # otherwise, if "drop_zero" / "drop_y_zero_x", compute predictions based on zero-filled
        # features which allows extrapolation.

    with timer("polars-ols w/ null_policy", precision=5):
        predictions = df.select(
            predictions=pl.col("y").least_squares.ols(
                pl.col("x1"), pl.col("x2"), null_policy=null_policy, mode="predictions"
            )
        )

    assert np.allclose(
        predictions.to_numpy().flatten(),
        predictions_expected.flatten(),
        rtol=1.0e-4,
        atol=1.0e-4,
        equal_nan=True,
    )

    # check residuals logic
    residuals_expected = y - predictions_expected  # no need to copy y
    residuals = df.select(
        residuals=pl.col("y").least_squares.ols(
            pl.col("x1"), pl.col("x2"), null_policy=null_policy, mode="residuals"
        )
    )
    assert np.allclose(
        residuals.to_numpy().flatten(), residuals_expected, rtol=1.0e-4, atol=1.0e-4, equal_nan=True
    )


def test_all_empty_data():
    df = pl.DataFrame(
        {
            "A": [None, 2, None, 4],
            "B": [1, None, 3, None],
        }
    )
    df = df.with_columns(
        pl.col("A")
        .least_squares.ols(
            pl.col("B"),
            mode="residuals",
            null_policy="drop",
            solve_method="svd",
        )
        .alias("residuals")
    )
    assert df["residuals"].is_null().all()


@pytest.mark.parametrize("n_features", (2, 10, 100, 1_000))
def test_fit_wide(n_features: int):
    df = _make_data(n_samples=10, n_features=n_features, scale=1.0e-4)
    features = [pl.col(f) for f in df.columns if f.startswith("x")]
    df = df.with_columns(
        # for p > k: OLS implementation will automatically use LAPACK SVD which should handle
        # over-determined problems
        pl.col("y").least_squares.ols(*features, mode="coefficients").alias("coef_ols"),
        # p >> k, cholesky of X.T @ X may fail: ridge will gracefully
        # fall back to LU with partial pivoting - which should handle over-determined problems
        pl.col("y")
        .least_squares.ridge(*features, mode="coefficients", alpha=1.0e-5)
        .alias("coef_ridge"),
        # lasso/elastic-net use coordinate-descent and so should work always
        pl.col("y")
        .least_squares.lasso(
            *features, mode="coefficients", alpha=1.0e-6, tol=1.0e-8, max_iter=3_000
        )
        .alias("coef_lasso"),
    )
    c_ols = df.select(
        pl.corr(
            pl.col("coef_ols").least_squares.predict(*features).alias("predictions"),
            pl.col("y"),
        )
    )
    c_ridge = df.select(
        pl.corr(
            pl.col("coef_ridge").least_squares.predict(*features).alias("predictions"),
            pl.col("y"),
        )
    )
    c_lasso = df.select(
        pl.corr(
            pl.col("coef_lasso").least_squares.predict(*features).alias("predictions"),
            pl.col("y"),
        )
    )
    assert c_ols.item() == pytest.approx(1.0, rel=1.0e-5, abs=1.0e-5)
    assert c_ridge.item() == pytest.approx(1.0, rel=1.0e-5, abs=1.0e-5)
    assert c_lasso.item() == pytest.approx(1.0, rel=1.0e-5, abs=1.0e-5)


@pytest.mark.parametrize(
    "n_features, solve_method",
    [
        (10, "svd"),  # n < k
        (99, "svd"),  # n = k
        (1_000, "svd"),  # n > k
        (90, "qr"),  # n < k
    ],
)
def test_fit_multi_collinear(n_features: int, solve_method: str):
    n_samples = 100
    # note that:
    # - only SVD solver (equivalent to lstsq) ensures minimum norm solution in case of collinearity
    last_feature_name = f"x{n_features}"
    multicollinear_feature_name = f"x{n_features + 1}"

    df = _make_data(n_samples=n_samples, n_features=n_features, scale=1.0e-4)
    df = df.with_columns((pl.col(last_feature_name) + 1.0e-12).alias(multicollinear_feature_name))

    features = [pl.col(f) for f in df.columns if f.startswith("x")]

    coef = (
        df.select(
            pl.col("y").least_squares.ols(
                *features, mode="coefficients", solve_method=solve_method, rcond=1.0e-16
            )  # machine precision rcond
        )
        .unnest("coefficients")
        .to_numpy()
        .flatten()
    )

    x, y = df.select(pl.all().exclude("y")).to_numpy(), df["y"].to_numpy()
    coef_expected = np.linalg.lstsq(x, y, rcond=1.0e-16)[0]  # similar to -1 (machine precision)

    if solve_method == "svd":
        assert np.allclose(
            coef, coef_expected, rtol=1.0e-2, atol=1.0e-2
        ), f"norm coef={np.linalg.norm(coef)}, norm expected: {np.linalg.norm(coef_expected)}"
        assert np.allclose(x @ coef, x @ coef_expected, rtol=1.0e-4, atol=1.0e-4)
    else:
        assert ~np.isnan(coef).any()
        # other methods will *not* same coefficients
        assert np.linalg.norm(coef) != np.linalg.norm(coef_expected)
        # but should not 'blow up' for QR
        assert np.allclose(x @ coef, x @ coef_expected, rtol=1.0e-4, atol=1.0e-4)


def test_coefficients_ols():
    df = _make_data()
    coef = (
        df.select(
            pl.col("y")
            .least_squares.from_formula("x1 + x2 -1", mode="coefficients")
            .alias("coefficients")
        )
        .unnest("coefficients")
        .to_numpy()[0]
    )
    assert np.allclose(coef, [1.0, 1.0], atol=1.0e-2, rtol=1.0e-2)


def test_coefficients_ols_groups():
    df = _make_data(n_groups=10)
    coef_group = (
        df.select(
            "group",
            pl.col("y")
            .least_squares.from_formula("x1 + x2 -1", mode="coefficients")
            .over("group")
            .alias("coefficients"),
        )
        .unique()
        .unnest("coefficients")
    )
    assert len(coef_group) == 10

    coef_group_1 = (
        df.filter(pl.col("group") == 1)
        .select(
            pl.col("y")
            .least_squares.from_formula("x1 + x2 -1", mode="coefficients")
            .alias("coefficients")
        )
        .unnest("coefficients")
    )
    assert np.allclose(coef_group.filter(pl.col("group") == 1).select("x1", "x2"), coef_group_1)


def test_coefficients_shape_broadcast():
    df = _make_data(n_samples=5_000, n_groups=10)
    assert df.select(
        pl.col("y")
        .least_squares.ols(pl.col("x1"), pl.col("x2"), mode="coefficients")
        .alias("coefficients")
    ).shape == (1, 1)

    assert df.with_columns(
        pl.col("y")
        .least_squares.ols(pl.col("x1"), pl.col("x2"), mode="coefficients")
        .alias("coefficients")
    ).shape == (5_000, 5)

    df_group = df.select(
        pl.col("y")
        .least_squares.ols(pl.col("x1"), pl.col("x2"), mode="coefficients")
        .over("group")
        .alias("coefficients"),
        "group",
    )
    assert df_group.shape == (5_000, 2)
    assert df_group.unique().shape == (10, 2)

    assert df.with_columns(
        pl.col("y")
        .least_squares.ols(pl.col("x1"), pl.col("x2"), mode="coefficients")
        .over("group")
        .alias("coefficients")
    ).shape == (5_000, 5)


def test_ols_residuals():
    df = _make_data()
    residuals = df.select(
        pl.col("y").least_squares.from_formula("x1 + x2 -1", mode="residuals")
    ).to_numpy()
    x, y = df.select("x1", "x2").to_numpy(), df["y"].to_numpy()
    coef = np.linalg.lstsq(x, y, rcond=None)[0]
    assert np.allclose(residuals.flatten(), y - x @ coef, rtol=1.0e-4, atol=1.0e-4)


def test_ols_intercept():
    df = _make_data()
    expr = compute_least_squares(pl.col("y"), pl.col("x1"), pl.col("x2"), add_intercept=True).alias(
        "predictions"
    )
    y_hat = df.select(expr).to_numpy().flatten()
    expected = smf.ols(formula="y ~ x1 + x2", data=df).fit().predict(df).to_numpy()
    assert np.allclose(y_hat, expected, atol=1.0e-4, rtol=1.0e-4)


def test_least_squares_from_formula():
    weights = np.random.uniform(0, 1, size=5_000)
    weights /= weights.mean()
    df = _make_data().with_columns(sample_weights=pl.lit(weights))

    expr = compute_least_squares_from_formula(
        "y ~ x1 + x2",  # patsy includes intercept by default
        sample_weights=pl.col("sample_weights"),
    ).alias("predictions")

    expected = (
        smf.wls("y ~ x1 + x2", data=df, weights=df["sample_weights"].to_numpy())
        .fit()
        .predict(df)
        .to_numpy()
    )
    assert np.allclose(df.select(expr).to_numpy().flatten(), expected, rtol=1.0e-4, atol=1.0e-4)


@pytest.mark.parametrize("solve_method", ("svd", "chol"))
def test_ridge(solve_method: str):
    df = _make_data()
    alpha: float = 0.01

    with timer(f"\nridge python benchmark - {solve_method}"):
        for _ in range(1_000):
            x = df.select("x1", "x2").to_numpy()
            y = df.select("y").to_numpy().flatten()
            if solve_method == "chol":
                coef_expected = np.linalg.solve((x.T @ x) + np.eye(x.shape[1]) * alpha, x.T @ y)
            elif solve_method == "svd":
                coef_expected = (
                    Ridge(fit_intercept=False, solver="svd", alpha=alpha).fit(x, y).coef_
                )
            else:
                raise ValueError()
            expected = x @ coef_expected

    with timer(f"ridge rust {solve_method}"):
        for _ in range(1_000):
            expr = compute_least_squares(
                pl.col("y"),
                pl.col("x1"),
                pl.col("x2"),
                ols_kwargs=OLSKwargs(alpha=alpha, solve_method=solve_method),
            ).alias("predictions")
            df = df.lazy().with_columns(expr).collect()
    assert np.allclose(df["predictions"].to_numpy(), expected, rtol=1.0e-4, atol=1.0e-4)


def test_wls():
    array = np.random.normal(size=(10_000, 2))
    df = pl.DataFrame(
        {
            "y": array.sum(axis=1)
            + np.hstack(
                [np.random.normal(size=8_000, scale=10.0), np.random.normal(size=2_000, scale=0.1)]
            ),
            "x1": array[:, 0],
            "x2": array[:, 1],
        }
    )

    weights = np.hstack([np.ones(8_000) * 1.0 / 10**2, np.ones(2_000) * 1.0 / 0.1**2])
    weights /= weights.mean()

    df = df.with_columns(sample_weight=weights)

    expr_wls = compute_least_squares(
        pl.col("y"),
        pl.col("x1"),
        pl.col("x2"),
        sample_weights=pl.col("sample_weight"),
    ).alias("predictions_wls")
    expr_ols = compute_least_squares(
        pl.col("y"),
        pl.col("x1"),
        pl.col("x2"),
    ).alias("predictions_ols")
    df = df.lazy().with_columns(expr_wls, expr_ols).collect()

    y_hat_wls = smf.wls(data=df, formula="y ~ x1 + x2 -1", weights=weights).fit().predict()
    y_hat_ols = smf.ols(data=df, formula="y ~ x1 + x2 -1").fit().predict()

    assert np.allclose(df["predictions_ols"].to_numpy(), y_hat_ols, rtol=1.0e-4, atol=1.0e-4)
    assert np.allclose(df["predictions_wls"].to_numpy(), y_hat_wls, rtol=1.0e-4, atol=1.0e-4)


def test_least_squares_namespace():
    df = _make_data().with_columns(sample_weight=pl.lit(1.0))
    alpha: float = 0.0
    ols = pl.col("y").least_squares.ols(pl.col("x1"), pl.col("x2")).alias("ols")
    ridge = pl.col("y").least_squares.ridge(pl.col("x1"), pl.col("x2"), alpha=alpha).alias("ridge")
    wls = (
        pl.col("y")
        .least_squares.wls(pl.col("x1"), pl.col("x2"), sample_weights=pl.col("sample_weight"))
        .alias("wls")
    )
    formula = pl.col("y").least_squares.from_formula("x1 + x2 - 1").alias("formula")
    df = df.lazy().select(ols, ridge, wls, formula).collect()

    # ensure all of the above are equivalent
    assert np.allclose(df.corr(), 1.0)


@pytest.mark.parametrize(
    "n_features,sparsity,alpha,solve_method",
    [
        (2, 0.0, 0.1, "cd"),
        (100, 0.5, 0.3, "cd"),
        (300, 0.5, 1.0, "cd"),
        (500, 0.9, 0.1, "cd"),
        (1_000, 0.9, 0.3, "cd_active_set"),
    ],
)
def test_elastic_net(n_features: int, sparsity: float, alpha: float, solve_method: SolveMethod):
    df = _make_data(n_features=n_features, sparsity=sparsity)
    features = df.select(pl.all().exclude("y")).columns

    with timer("\nelastic net sklearn"):
        mdl = ElasticNet(fit_intercept=False, alpha=alpha, l1_ratio=0.5, max_iter=1_000, tol=0.0001)
        x, y = df.select(*features), df.select("y")
        mdl.fit(x, y)
        predictions_1 = mdl.predict(x).flatten()

    with timer("elastic net polars-ols"):
        predictions_2 = (
            df.lazy()
            .select(
                pl.col("y").least_squares.elastic_net(
                    *features,
                    mode="predictions",
                    l1_ratio=0.5,
                    alpha=alpha,
                    max_iter=1_000,
                    tol=0.0001,
                    solve_method=solve_method,
                )
            )
            .collect()
            .to_numpy()
            .flatten()
        )
    assert np.allclose(predictions_1, predictions_2, rtol=1.0e-4, atol=1.0e-4)


def test_elastic_net_non_negative():
    df = _make_data()
    mdl = ElasticNet(
        fit_intercept=False, alpha=0.1, l1_ratio=0.5, max_iter=1_000, tol=0.0001, positive=True
    )
    mdl.fit(df.select(pl.col("x1"), -pl.col("x2")), df.select("y"))

    coef = (
        df.lazy()
        .select(
            pl.col("y")
            .least_squares.elastic_net(
                pl.col("x1"),
                -pl.col("x2"),
                mode="coefficients",
                l1_ratio=0.5,
                alpha=0.1,
                max_iter=1_000,
                tol=0.0001,
                positive=True,
            )
            .alias("coefficients")
        )
        .unnest("coefficients")
        .collect()
        .to_numpy()
        .flatten()
    )
    assert np.allclose(mdl.coef_, coef, rtol=1.0e-4, atol=1.0e-4)


def test_recursive_least_squares():
    df = _make_data()
    rng = np.random.default_rng(0)

    def insert_nulls(val):
        return None if rng.random() < 0.1 else val

    df = df.with_columns(
        *(pl.col(c).map_elements(insert_nulls, return_dtype=pl.Float64) for c in df.columns)
    )

    # expanding OLS
    coef_rls = (
        df.lazy()
        .select(
            pl.col("y")
            .least_squares.rls(
                pl.col("x1"),
                pl.col("x2"),
                mode="coefficients",
                # equivalent to expanding window (no forgetting)
                half_life=None,
                # arbitrarily weak L2 (diffuse) prior
                initial_state_covariance=1_000_000.0,
                null_policy="drop",
            )
            .alias("coefficients")
        )
        .unnest("coefficients")
        .collect()
        .to_numpy()
    )

    # full sample OLS
    coef_ols = (
        df.lazy()
        .select(
            pl.col("y")
            .least_squares.ols(pl.col("x1"), pl.col("x2"), mode="coefficients", null_policy="drop")
            .alias("coefficients")
        )
        .unnest("coefficients")
        .collect()
        .to_numpy()
        .flatten()
    )

    # the two models should be equivalent in full sample
    assert np.allclose(coef_rls[-1], coef_ols, rtol=1.0e-4, atol=1.0e-4)


def test_recursive_least_squares_prior():
    df = _make_data()

    # equivalent to expanding OLS w/ L2 prior towards [0.5, 0.5]
    coef_rls_prior = (
        df.lazy()
        .select(
            pl.col("y")
            .least_squares.rls(
                pl.col("x1"),
                pl.col("x2"),
                mode="coefficients",
                # equivalent to expanding window (no forgetting)
                half_life=None,
                initial_state_covariance=1.0e-6,  # arbitrarily strong L2 prior
                initial_state_mean=[0.25, 0.25],  # custom prior
            )
            .alias("coefficients")
        )
        .unnest("coefficients")
        .collect()
        .to_numpy()
    )

    # given few samples and strong prior strength, the coefficients are nearly
    # identical to the prior
    assert np.allclose(coef_rls_prior[0], [0.25, 0.25], rtol=1.0e-3, atol=1.0e-3)
    assert np.allclose(coef_rls_prior[10], [0.25, 0.25], rtol=1.0e-3, atol=1.0e-3)

    # as number of samples seen grows, the coefficients start to diverge from prior
    # & eventually converge to ground truth.
    assert not np.allclose(coef_rls_prior[-1], [0.5, 0.5], rtol=1.0e-4, atol=1.0e-4)


@pytest.mark.parametrize(
    "window_size,min_periods,use_woodbury",
    [
        (2, 2, False),
        (10, 2, False),
        (10, 2, True),
        (63, 5, False),
        (252, 5, False),
        (252, 5, True),
    ],
)
def test_rolling_least_squares(window_size: int, min_periods: int, use_woodbury: bool):
    df = _make_data(n_samples=1_000, add_missing=True, missing_columns=("y",))

    with timer("\nrolling ols"):
        coef_rolling = (
            df.lazy()
            .select(
                pl.col("y")
                .least_squares.rolling_ols(
                    pl.col("x1"),
                    pl.col("x2"),
                    mode="coefficients",
                    window_size=window_size,
                    min_periods=min_periods,
                    use_woodbury=use_woodbury,
                    null_policy="drop_window",  # equivalent to mode='missing' in statsmodels
                )
                .alias("coefficients")
            )
            .unnest("coefficients")
            .collect()
        )
    with timer("rolling ols statsmodels"):
        mdl = RollingOLS(
            df["y"].to_numpy(),
            df[["x1", "x2"]].to_numpy(),
            window=window_size,
            min_nobs=min_periods,
            expanding=True,
            missing="drop",
        ).fit()
        coef_statsmodels = (
            pl.DataFrame(mdl.params, schema=["x1", "x2"])
            .fill_nan(None)
            .with_columns(pl.all().forward_fill())
        )

    assert np.allclose(
        coef_rolling.to_numpy(),
        coef_statsmodels.to_numpy(),
        rtol=1.0e-3,
        atol=1.0e-3,
        equal_nan=True,
    )


@pytest.mark.parametrize(
    "min_periods,expected",
    [
        (999, 2),
        (1_000, 1),
        (1_001, 0),
    ],
)
def test_rolling_ols_insufficient_data(min_periods: int, expected: int):
    df = _make_data(n_samples=1_000)

    with timer("\nrolling ols"):
        coef_rolling = (
            df.lazy()
            .select(
                pl.col("y")
                .least_squares.rolling_ols(
                    pl.col("x1"),
                    pl.col("x2"),
                    mode="coefficients",
                    window_size=2_000,
                    min_periods=min_periods,
                    use_woodbury=False,
                    null_policy="drop_window",  # equivalent to mode='missing' in statsmodels
                )
                .alias("coefficients")
            )
            .unnest("coefficients")
            .collect()
        )

    assert coef_rolling.count().max_horizontal().item() == expected


@pytest.mark.parametrize("window_size", (21, 252))
def test_rolling_window_drop(window_size: int):
    df = _make_data(n_samples=1_000, add_missing=True, missing_columns=("y",)).with_row_index()
    df_drop = df.drop_nulls()

    df_merged = (
        df_drop.lazy()
        .select(
            "index",
            pl.col("y")
            .least_squares.rolling_ols(
                "x1", "x2", window_size=window_size, mode="predictions", null_policy="drop"
            )
            .alias("predictions_1"),
        )
        .join_asof(
            df.lazy().select(
                "index",
                pl.col("y")
                .least_squares.rolling_ols(
                    "x1",
                    "x2",
                    window_size=window_size,
                    mode="predictions",
                    null_policy="drop",
                )
                .alias("predictions_2"),
            ),
            on="index",
        )
    ).collect()

    assert np.allclose(df_merged["predictions_1"], df_merged["predictions_2"], equal_nan=True)


def test_moving_window_regressions_over():
    df = _make_data(n_groups=10)

    df = (
        (
            df.lazy().select(
                "group",
                pl.col("y")
                .least_squares.rolling_ols(
                    pl.col("x1"),
                    pl.col("x2"),
                    mode="coefficients",
                    window_size=1_000_000,
                    min_periods=2,
                    # larger than data window size equivalent to expanding window
                )
                .over("group")
                .alias("coef_rolling_ols_group"),
                pl.col("y")
                .least_squares.rls(
                    pl.col("x1"),
                    pl.col("x2"),
                    half_life=None,
                    initial_state_covariance=1.0e6,
                    mode="coefficients",
                    # no forgetting factor + diffuse prior equivalent to expanding window
                )
                .over("group")
                .alias("coef_rls_group"),
                pl.col("y")
                .least_squares.ols(
                    pl.col("x1"),
                    pl.col("x2"),
                    mode="coefficients",
                )
                .over("group")
                .alias("coef_ols_group"),  # full sample OLS per group
            )
        )
        .collect()
        .rechunk()
    )

    # As of the last sample per group: RLS & rolling regression should behave identically
    # to full sample OLS per group (the way they were set up above)
    df_last = df.group_by("group").last()

    assert np.allclose(
        df_last.unnest("coef_ols_group").select("x1", "x2"),
        df_last.unnest("coef_rolling_ols_group").select("x1", "x2"),
    )
    assert np.allclose(
        df_last.unnest("coef_ols_group").select("x1", "x2"),
        df_last.unnest("coef_rls_group").select("x1", "x2"),
        rtol=1.0e-4,
        atol=1.0e-4,
    )


def test_predict():
    df = _make_data(n_groups=1)
    df_test = _make_data(n_samples=20, n_groups=1).drop("y")

    # estimate coefficients
    df_coefficients = (
        df.lazy()
        .select(
            "group",
            pl.col("y")
            .least_squares.ols(pl.col("x1"), pl.col("x2"), mode="coefficients")
            .over("group")
            .alias("coefficients"),
        )
        .unique()
    )

    # compute predictions as a rust expression
    predictions = (
        df_test.lazy()
        .join(df_coefficients, on="group")
        .select(
            pl.col("coefficients").least_squares.predict(
                pl.col("x1"),
                pl.col("x2"),
                name="predictions",
                null_policy="zero",
            )
        )
        .collect()
        .to_numpy()
        .flatten()
    )

    # compare to predictions computed as dot product of train coefficients with test data
    expected = (
        df_test.drop("group")
        .to_numpy()
        .dot(df_coefficients.drop("group").unnest("coefficients").collect().to_numpy().T)
        .flatten()
    )
    assert np.allclose(predictions, expected)


def test_predict_formula():
    df = _make_data()
    df = (
        df.lazy()
        .with_columns(
            coefficients=pl.col("y").least_squares.from_formula("x1 + x2", mode="coefficients"),
            predictions_1=pl.col("y").least_squares.from_formula("x1 + x2", mode="predictions"),
        )
        .with_columns(
            predictions_2=pl.col("coefficients").least_squares.predict_from_formula("x1 + x2")
        )
    ).collect()
    assert np.allclose(df["predictions_1"], df["predictions_2"])


def test_predict_complex():
    df = _make_data(n_groups=10)
    df = (
        df.lazy()
        .with_columns(
            predictions_1=pl.col("y")
            .least_squares.rls(pl.col("x1"), pl.col("x2"), mode="predictions")
            .over("group"),
            coefficients=pl.col("y")
            .least_squares.rls(pl.col("x1"), pl.col("x2"), mode="coefficients")
            .over("group"),
        )
        .with_columns(
            predictions_2=pl.col("coefficients").least_squares.predict(pl.col("x1"), pl.col("x2"))
        )
        .collect()
    )
    assert np.allclose(df["predictions_1"], df["predictions_2"])



import numpy as np
import polars as pl
import pyperf
import scipy
import statsmodels.formula.api as smf
from sklearn.linear_model import ElasticNet, LinearRegression, Ridge
from statsmodels.regression.recursive_ls import RecursiveLS
from statsmodels.regression.rolling import RollingOLS

import polars_ols as pls  # import package to register the .least_squares namespace
from polars_ols.least_squares import SolveMethod


def _make_data(
    n_samples: int = 2_000,
    n_features: int = 5,
    sparsity: float = 0.5,
    n_targets: int = 1,
) -> pl.DataFrame:
    x = np.random.normal(size=(n_samples, n_features))  # N x K
    features = pl.DataFrame(data=x, schema=[f"x{i + 1}" for i in range(n_features)])
    # make single output
    if n_targets == 1:
        eps = np.random.normal(size=n_samples, scale=0.1)
        return features.with_columns(
            y=pl.lit(x[:, : int(n_features * (1.0 - sparsity))].sum(1) + eps)
        )
    # make multi-target
    else:
        # create a random linear kernel
        w = np.random.normal(size=(x.shape[1], n_targets))
        eps = np.random.normal(size=(n_samples, n_targets), scale=0.1)
        sparsity_mask = np.random.choice([True, False], size=w.shape, p=[1 - sparsity, sparsity])
        w *= sparsity_mask
        y = x @ w + eps
        return features.with_columns(
            pl.lit(y).list.to_struct(fields=[f"y{i}" for i in range(n_targets)]).alias("y")
        )


def benchmark_least_squares(data: pl.DataFrame, solve_method: SolveMethod = "qr"):
    return (
        data.lazy()
        .with_columns(
            (
                pl.col("y")
                .least_squares.ols(pl.all().exclude("y"), solve_method=solve_method)
                .alias("predictions")
            )
        )
        .collect()
    )


def benchmark_least_squares_numpy_svd(data: pl.DataFrame):
    y, x = data.select("y").to_numpy().flatten(), data.select(pl.all().exclude("y")).to_numpy()
    coef = np.linalg.lstsq(x, y, rcond=None)[0]
    return data.lazy().with_columns(predictions=pl.lit(x @ coef).alias("predictions")).collect()


def benchmark_least_squares_numpy_qr(data: pl.DataFrame):
    y, x = data.select("y").to_numpy().flatten(), data.select(pl.all().exclude("y")).to_numpy()
    q, r = np.linalg.qr(x)
    # make use of the fact that r is upper triangular --> use back substitution (solve_triangular)
    coef = scipy.linalg.solve_triangular(r, q.T @ y, check_finite=False)
    return data.lazy().with_columns(predictions=pl.lit(x @ coef).alias("predictions")).collect()


def benchmark_ridge(data: pl.DataFrame, solve_method: SolveMethod = "chol"):
    return (
        data.lazy()
        .with_columns(
            pl.col("y")
            .least_squares.ridge(
                *[pl.col(c) for c in data.columns if c != "y"],
                alpha=0.0001,
                solve_method=solve_method,
            )
            .alias("predictions")
        )
        .collect()
    )


def benchmark_ridge_sklearn(data: pl.DataFrame, solve_method: str = "cholesky"):
    alpha: float = 0.0001
    y, x = data.select("y").to_numpy().flatten(), data.select(pl.all().exclude("y")).to_numpy()
    mdl = Ridge(fit_intercept=False, alpha=alpha, solver=solve_method).fit(x, y)
    return (
        data.lazy().with_columns(predictions=pl.lit(x @ mdl.coef_).alias("predictions")).collect()
    )


def benchmark_wls_from_formula(data: pl.DataFrame):
    return (
        data.lazy()
        .with_columns(pl.lit(1.0).alias("sample_weights"))
        .with_columns(
            pl.col("y")
            .least_squares.from_formula(
                " + ".join(c for c in data.columns if "x" in c),
                sample_weights=pl.col("sample_weights"),
            )
            .alias("predictions")
        )
    ).collect()


def benchmark_wls_from_formula_statsmodels(data: pl.DataFrame):
    x_cols = " + ".join(c for c in data.columns if "x" in c)
    predictions = (
        smf.wls(
            data=data,
            formula=f"y ~ {x_cols}",
            weights=np.ones(len(data), dtype="float32"),
        )
        .fit()
        .predict()
    )
    return data.lazy().with_columns(predictions=pl.lit(predictions)).collect()


def benchmark_elastic_net(data: pl.DataFrame, solve_method: SolveMethod = "cd_active_set"):
    return (
        data.lazy()
        .with_columns(
            pl.col("y").least_squares.elastic_net(
                *[pl.col(c) for c in data.columns if c != "y"],
                alpha=0.1,
                l1_ratio=0.5,  # same as sklearn default setting
                max_iter=1_000,  # same as sklearn default setting
                tol=1.0e-4,  # same as sklearn default setting
                solve_method=solve_method,
            )
        )
        .collect()
    )


def benchmark_elastic_net_sklearn(data: pl.DataFrame):
    mdl = ElasticNet(fit_intercept=False, alpha=0.1, l1_ratio=0.5, max_iter=1_000)
    y, x = data.select("y").to_numpy().flatten(), data.select(pl.all().exclude("y")).to_numpy()
    mdl.fit(x, y)
    return data.lazy().with_columns(predictions=pl.lit(mdl.predict(x))).collect()


def benchmark_recursive_least_squares(data: pl.DataFrame):
    return (
        data.lazy()
        .with_columns(
            pl.col("y").least_squares.rls(
                *[pl.col(c) for c in data.columns if c != "y"],
                half_life=252,
            )
        )
        .collect()
    )


def benchmark_rolling_least_squares(data: pl.DataFrame):
    features = [pl.col(c) for c in data.columns if c != "y"]
    return (
        data.lazy()
        .with_columns(
            pl.col("y").least_squares.rolling_ols(
                *features,
                window_size=252,
                min_periods=len(features),
                null_policy="drop_window",
            )
        )
        .collect()
    )


def benchmark_rolling_least_squares_statsmodels(data: pl.DataFrame):
    x = data.select(pl.all().exclude("y")).to_numpy()
    res = RollingOLS(df["y"].to_numpy(), x, window=252, min_nobs=x.shape[1], expanding=True).fit(
        params_only=True
    )
    return data.lazy().with_columns(predictions=pl.lit((res.params * x).sum(1))).collect()


def benchmark_recursive_least_squares_statsmodels(data: pl.DataFrame):
    x = data.select(pl.all().exclude("y")).to_numpy()
    res = RecursiveLS(
        df["y"].to_numpy(),
        x,
    ).fit()
    return data.lazy().with_columns(predictions=pl.lit((res.params * x).sum(1))).collect()


def benchmark_multi_target(data: pl.DataFrame):
    return (
        data.lazy()
        .with_columns(
            predictions=pl.col("y").least_squares.multi_target_ols(
                pl.all().exclude("y"),
                mode="predictions",
            )
        )
        .collect()
    )


def benchmark_multi_target_sklearn(data: pl.DataFrame):
    y, x = data.select("y").unnest("y").to_numpy(), data.select(pl.all().exclude("y")).to_numpy()
    mdl = LinearRegression(fit_intercept=False)
    mdl.fit(x, y)
    return data.with_columns(
        pl.lit(mdl.predict(x)).list.to_struct(fields=[f"y{i}" for i in range(y.shape[1])])
    )


if __name__ == "__main__":
    # example: python tests/benchmark.py --quiet --rigorous
    # we run the benchmarks in python (as opposed to rust) so that overhead of pyO3 is included
    df = _make_data(n_features=100, n_samples=10_000)
    df_multi_target = _make_data(n_features=100, n_targets=20, n_samples=10_000)

    runner = pyperf.Runner()
    runner.bench_func("benchmark_least_squares_svd", benchmark_least_squares, df, "svd")
    runner.bench_func("benchmark_least_squares_qr", benchmark_least_squares, df, "qr")
    runner.bench_func("benchmark_multi_target", benchmark_multi_target, df_multi_target)

    runner.bench_func("benchmark_least_squares_numpy_svd", benchmark_least_squares_numpy_svd, df)
    runner.bench_func("benchmark_least_squares_numpy_qr", benchmark_least_squares_numpy_qr, df)
    runner.bench_func("benchmark_multi_target_sklearn", benchmark_multi_target, df_multi_target)

    # runner.bench_func("benchmark_ridge_cholesky", benchmark_ridge, df, "chol")
    # runner.bench_func("benchmark_ridge_svd", benchmark_ridge, df, "svd")
    # runner.bench_func("benchmark_wls_from_formula", benchmark_wls_from_formula, df)
    # runner.bench_func("benchmark_elastic_net", benchmark_elastic_net, df, "cd")
    # runner.bench_func(
    #     "benchmark_elastic_net_active_set", benchmark_elastic_net, df, "cd_active_set"
    # )
    # runner.bench_func("benchmark_recursive_least_squares", benchmark_recursive_least_squares, df)
    # runner.bench_func("benchmark_rolling_least_squares", benchmark_rolling_least_squares, df)

    # runner.bench_func("benchmark_ridge_sklearn_cholesky", benchmark_ridge_sklearn, df, "cholesky")
    # runner.bench_func("benchmark_ridge_sklearn_svd", benchmark_ridge_sklearn, df, "svd")
    # runner.bench_func(
    #     "benchmark_wls_from_formula_statsmodels", benchmark_wls_from_formula_statsmodels, df
    # )
    # runner.bench_func("benchmark_elastic_net_sklearn", benchmark_elastic_net_sklearn, df)
    # runner.bench_func(
    #     "benchmark_recursive_least_squares_statsmodels",
    #     benchmark_recursive_least_squares_statsmodels,
    #     df,
    # )
    # runner.bench_func(
    #     "benchmark_rolling_least_squares_statsmodels",
    #     benchmark_rolling_least_squares_statsmodels,
    #     df,
    # )



from __future__ import annotations

from typing import List, Optional, Union

import polars as pl

from polars_ols.least_squares import (
    NullPolicy,
    OLSKwargs,
    OutputMode,
    RLSKwargs,
    RollingKwargs,
    SolveMethod,
    compute_least_squares,
    compute_least_squares_from_formula,
    compute_multi_target_least_squares,
    compute_recursive_least_squares,
    compute_rolling_least_squares,
    predict,
)
from polars_ols.utils import build_expressions_from_patsy_formula

__all__ = [
    "compute_least_squares",
    "compute_recursive_least_squares",
    "compute_rolling_least_squares",
    "compute_multi_target_least_squares",
    "compute_least_squares_from_formula",
    "LeastSquares",
]

ExprOrStr = Union[pl.Expr, str]


@pl.api.register_expr_namespace("least_squares")
class LeastSquares:
    """Registers the `.least_squares` namespace and provide entry points for the various models
     supported by this extension package.

     The below are parameters common to all models:
        - `sample_weights`: Optional expression representing sample weights.
        - `add_intercept`: Whether to add an intercept column.
        - `mode`: Mode of operation ("predictions", "residuals", "coefficients").
        - `null_policy`: Strategy for handling missing data, it can be:
            "ignore": does no null handling - use this option if nulls are already handled upstream.
            "zero": simply zero fills nulls in both targets & features prior to fitting.
            "drop": drops any rows (across targets or features) which have nulls prior to computing
                    coefficients. For non moving-window models, it then masks associated predictions
                    with nulls. For moving-window models coefficients are forward-filled.
            "drop_y_zero_x": Similar to "drop", but only rows with null targets are masked.
                             If features contain remaining nulls, it fills them with zeros.
                             This option allows extrapolation.
            "drop_zero": Masks rows with null targets or features similar in fitting coefficients.
                         similar to "drop". However, for predictions it uses zero filled features
                         dotted with those coefficients instead of masking with null.
                         This option allows extrapolation.

    Remaining parameters are model specific, see `OLSKwargs`, `RLSKwargs`, & `RollingKwargs`
    dataclasses for details.
    """

    def __init__(self, expr: pl.Expr):
        self._expr = expr

    def least_squares(
        self,
        *features: ExprOrStr,
        sample_weights: Optional[ExprOrStr] = None,
        add_intercept: bool = False,
        mode: OutputMode = "predictions",
        null_policy: NullPolicy = "ignore",
        solve_method: Optional[SolveMethod] = None,
        multi_target: bool = False,
        **ols_kwargs,
    ) -> pl.Expr:
        """Perform least squares regression.

        :param features: Variable number of feature expressions.
        :param sample_weights: Optional expression representing sample weights.
        :param add_intercept: Whether to add an intercept column.
        :param mode: Mode of operation ("predictions", "residuals", "coefficients").
        :param null_policy: Strategy for handling missing data.
        :param solve_method: Algorithm used for computing least squares solution.
                             Defaults to None, where a recommended default method is chosen based
                             on parametrization of the least-squares problem.
                             It can be one of ("qr", "svd", "chol", "lu", "cd").
        :param multi_target: boolean indicating if the target expression is multi-target struct.
        :param ols_kwargs: Additional, optional, model specific kwargs. See OLSKwargs.
        """
        ols_func = compute_least_squares if not multi_target else compute_multi_target_least_squares
        return ols_func(
            self._expr,
            *features,
            sample_weights=sample_weights,
            add_intercept=add_intercept,
            mode=mode,
            ols_kwargs=OLSKwargs(null_policy=null_policy, solve_method=solve_method, **ols_kwargs),
        )

    def ols(self, *features: ExprOrStr, **kwargs) -> pl.Expr:
        """Performs ordinary least squares. Alias for `least_squares`."""
        return self.least_squares(*features, **kwargs)

    def multi_target_ols(self, *features: ExprOrStr, **kwargs) -> pl.Expr:
        return self.least_squares(*features, multi_target=True, **kwargs)

    def wls(self, *features: ExprOrStr, sample_weights: ExprOrStr, **kwargs) -> pl.Expr:
        """Performs weighted least squares. Alias for `least_squares`.

        :param features: Variable number of feature expressions.
        :param sample_weights: optional expression representing sample weights.
        :param kwargs: additional kwargs passed to `least_squares` function.
        """
        return self.least_squares(*features, sample_weights=sample_weights, **kwargs)

    def ridge(self, *features: ExprOrStr, alpha: float, **kwargs) -> pl.Expr:
        """Performs ridge regression. Alias for `least_squares`.

        :param features: Variable number of feature expressions.
        :param alpha: L2 regularization parameter. Must be non-negative.
        :param kwargs: additional kwargs passed to `least_squares` function.
        """
        return self.least_squares(*features, alpha=alpha, l1_ratio=0.0, **kwargs)

    def lasso(self, *features: ExprOrStr, alpha: float, **kwargs) -> pl.Expr:
        """Performs lasso regression. Alias for `least_squares`.

        :param features: Variable number of feature expressions.
        :param alpha: L1 regularization parameter. Must be non-negative.
        :param kwargs: additional kwargs passed to `least_squares` function.
        """
        return self.least_squares(*features, alpha=alpha, l1_ratio=1.0, **kwargs)

    def elastic_net(
        self,
        *features: ExprOrStr,
        alpha: float,
        l1_ratio: float = 0.5,
        positive: bool = False,
        **kwargs,
    ):
        """Performs lasso regression. Alias for `least_squares`.

        The implementation uses cyclic coordinate descent.

        :param features: Variable number of feature expressions.
        :param alpha: Regularization strength parameter. Must be non-negative.
        :param l1_ratio: mixing parameter for elastic net regularization. Must be in [0., 1.]
                         (0. for Ridge, 1. for LASSO).
        :param positive: if set, enforces non-negativity constraints on coefficients.
        :param kwargs: additional kwargs passed to `least_squares` function.
        """
        return self.least_squares(
            *features, alpha=alpha, l1_ratio=l1_ratio, positive=positive, **kwargs
        )

    def rls(
        self,
        *features: ExprOrStr,
        sample_weights: Optional[ExprOrStr] = None,
        add_intercept: bool = False,
        mode: OutputMode = "predictions",
        null_policy: NullPolicy = "drop",
        half_life: Optional[float] = None,
        initial_state_covariance: Optional[float] = 10.0,
        initial_state_mean: Union[Optional[List[float], float]] = None,
    ):
        """Performs recursive least squares estimation.

        This model incrementally updates the least squares coefficients per observation
        in an efficient manner.

        General parameters:
        :param features: Variable number of feature expressions.
        :param sample_weights: Optional expression representing sample weights.
        :param add_intercept: Whether to add an intercept column.
        :param mode: Mode of operation ("predictions", "residuals", "coefficients").
        :param null_policy: Strategy for handling missing data. Recursive least squares model
                            will always forward fill coefficients in cased of masked observations.

        Model specific parameters:
        :param half_life: Half-life parameter for exponential forgetting. Defaults to None, which is
                          equivalent to expanding window least-squares (no forgetting).
        :param initial_state_mean: Denotes a prior for the mean value of the coefficients.
                          It can be either a scalar or a list. Defaults to None (no prior).
        :param initial_state_covariance: Scalar which denotes a prior of the uncertainty (variance)
                          around `initial_state_mean`. It can be thought of as inversely
                          proportional to the strength of an L2 penalty in a ridge regression.
                          Defaults to 10, intended to be a weak 'diffuse' prior.
        """
        return compute_recursive_least_squares(
            self._expr,
            *features,
            sample_weights=sample_weights,
            add_intercept=add_intercept,
            mode=mode,
            rls_kwargs=RLSKwargs(
                null_policy=null_policy,
                half_life=half_life,
                initial_state_mean=initial_state_mean,
                initial_state_covariance=initial_state_covariance,
            ),
        )

    def rolling_ols(
        self,
        *features: ExprOrStr,
        window_size: int,
        sample_weights: Optional[ExprOrStr] = None,
        add_intercept: bool = False,
        mode: OutputMode = "predictions",
        null_policy: NullPolicy = "drop",
        min_periods: Optional[int] = None,
        use_woodbury: Optional[bool] = None,
        alpha: Optional[float] = None,
    ):
        """Performs rolling window least squares estimation by incrementally updating its state.

        General parameters:
        :param features: Variable number of feature expressions.
        :param sample_weights: Optional expression representing sample weights.
        :param add_intercept: Whether to add an intercept column.
        :param mode: Mode of operation ("predictions", "residuals", "coefficients").
        :param null_policy: Strategy for handling missing data. See `NullPolicy` for all options.
            Specific to the rolling window context:
                - "drop_window": This policy is specific to rolling ols. For every fixed rolling
                 window of data: drop rows with null values and only use remaining observations
                 for estimation. If number of valid observations per window doesn't exceed
                 min_periods: forward fill last previous coefficients. Note that this behavior is
                 similar to statsmodels RollingOLS' missing='drop' behaviour.
                - "drop": Logically equivalent to dropping any rows with null rows,
                computing rolling window on that data, then re-aligning with forward fill to the
                original data shape. I.e. it always operates on the last `window_size`
                valid observations.

        Model specific parameters:
        :param window_size: The size of the rolling window.
        :param min_periods: The minimum number of observations required to produce estimates.
        :param use_woodbury: Whether to use the Woodbury matrix identity for faster computation,
                             this allows updating inv(XTX) directly avoiding an expensive inversion
                             if number of features is large.
                             Defaults to None: which triggers Woodbury updates if num_features > 60.
        :param alpha: L2 Regularization strength. Default is 0.0.
        """
        return compute_rolling_least_squares(
            self._expr,
            *features,
            sample_weights=sample_weights,
            add_intercept=add_intercept,
            mode=mode,
            rolling_kwargs=RollingKwargs(
                window_size=window_size,
                min_periods=min_periods,
                use_woodbury=use_woodbury,
                alpha=alpha,
                null_policy=null_policy,
            ),
        )

    def expanding_ols(self, *features: ExprOrStr, **kwargs):
        return self.rls(*features, half_life=None, **kwargs)

    def from_formula(self, formula: str, **kwargs) -> pl.Expr:
        features, add_intercept = build_expressions_from_patsy_formula(
            formula, include_dependent_variable=False
        )
        if kwargs.get("half_life"):
            return self.rls(*features, add_intercept=add_intercept, **kwargs)
        elif kwargs.get("window_size"):
            return self.rolling_ols(*features, add_intercept=add_intercept, **kwargs)
        else:
            return self.least_squares(*features, add_intercept=add_intercept, **kwargs)

    def predict(
        self,
        *features: ExprOrStr,
        name: Optional[str] = None,
        add_intercept: bool = False,
        null_policy: NullPolicy = "zero",
    ) -> pl.Expr:
        return predict(
            self._expr,
            *features,
            add_intercept=add_intercept,
            name=name,
            null_policy=null_policy,
        )

    def predict_from_formula(self, formula: str, name: Optional[str] = None) -> pl.Expr:
        features, add_intercept = build_expressions_from_patsy_formula(
            formula, include_dependent_variable=False
        )
        has_const = any(f.meta.output_name == "const" for f in features)
        add_intercept &= not has_const
        return self.predict(*features, name=name, add_intercept=add_intercept)
